/* vim: set sw=4 sts=4 et foldmethod=syntax : */

#ifndef GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_GRAPH_TRAITS_HH
#define GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_GRAPH_TRAITS_HH 1

#include "formats/input_graph.hh"

auto is_simple_clique(const InputGraph & graph) -> bool;

#endif
