/* vim: set sw=4 sts=4 et foldmethod=syntax : */

#ifndef GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_VERTEX_TO_VERTEX_MAPPING_HH
#define GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_VERTEX_TO_VERTEX_MAPPING_HH 1

#include <map>

using VertexToVertexMapping = std::map<int, int>;

#endif
