/* vim: set sw=4 sts=4 et foldmethod=syntax : */

#ifndef GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_THREAD_UTILS_HH
#define GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_THREAD_UTILS_HH 1

auto how_many_threads(unsigned n) -> unsigned;

#endif
