/* vim: set sw=4 sts=4 et foldmethod=syntax : */

#ifndef GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_LOOOONG_HH
#define GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_LOOOONG_HH 1

#include <boost/multiprecision/cpp_int.hpp>

using loooong = boost::multiprecision::cpp_int;

#endif
